@@ -3,25 +3,205 @@ if numexpr.__version__ < 2.0:
     raise ImportError('Please update numexpr to a newer version > 2.0.')
 
 import numpy as np
-from .decorators import real, square, idempotent, inplace, separable
+from .decorators import idempotent, inplace, real, separable, square
 from .core import (Operator, BlockColumnOperator, CompositionOperator,
-                   IdentityOperator, ReductionOperator, ReverseOperatorFactory)
-from .utils import operation_assignment, operation_symbol, strenum, tointtuple
+                   IdentityOperator, ReductionOperator)
+from .linear import DegreesOperator, RadiansOperator
+from .utils import (
+    operation_assignment, operation_symbol, pi, strenum, tointtuple)
 from .utils.ufuncs import hard_thresholding, soft_thresholding
 
-__all__ = ['ClipOperator',
+__all__ = ['Cartesian2SphericalOperator',
+           'ClipOperator',
            'HardThresholdingOperator',
            'MaxOperator',
            'MinOperator',
            'MinMaxOperator',
            'MaximumOperator',
            'MinimumOperator',
+           'NormalizeOperator',
            'NumexprOperator',
            'ProductOperator',
            'RoundOperator',
+           'SoftThresholdingOperator',
+           'Spherical2CartesianOperator',
            'To1dOperator',
-           'ToNdOperator',
-           'SoftThresholdingOperator']
+           'ToNdOperator']
+
+
+@real
+class _CartesianSpherical(Operator):
+    CONVENTIONS = ('zenith,azimuth',
+                   'azimuth,zenith',
+                   'elevation,azimuth',
+                   'azimuth,elevation')
+
+    def __init__(self, convention, dtype=float, **keywords):
+        if not isinstance(convention, str):
+            raise TypeError("The input convention '{0}' is not a string.".
+                            format(convention))
+        convention_ = convention.replace(' ', '').lower()
+        if convention_ not in self.CONVENTIONS:
+            raise ValueError(
+                "Invalid spherical convention '{0}'. Expected values are {1}.".
+                format(convention, strenum(self.CONVENTIONS)))
+        self.convention = convention_
+        Operator.__init__(self, dtype=dtype, **keywords)
+
+    @staticmethod
+    def _reshapecartesian(shape):
+        return shape[:-1] + (2,)
+
+    @staticmethod
+    def _reshapespherical(shape):
+        return shape[:-1] + (3,)
+
+    @staticmethod
+    def _validatecartesian(shape):
+        if len(shape) == 0 or shape[-1] != 3:
+            raise ValueError('Invalid cartesian shape.')
+
+    @staticmethod
+    def _validatespherical(shape):
+        if len(shape) == 0 or shape[-1] != 2:
+            raise ValueError('Invalid spherical shape.')
+
+    @staticmethod
+    def _rule_identity(s, o):
+        if s.convention == o.convention:
+            return IdentityOperator()
+
+
+class Cartesian2SphericalOperator(_CartesianSpherical):
+    """
+    Convert cartesian unit vectors into spherical coordinates in radians
+    or degrees.
+
+    The spherical coordinate system is defined by:
+       - the zenith direction of coordinate (0, 0, 1)
+       - the azimuthal reference of coordinate (1, 0, 0)
+       - the azimuth signedness: it is counted positively from the X axis
+    to the Y axis.
+
+    The last dimension of the operator's output is 2 and it encodes
+    the two spherical angles. Four conventions define what these angles are:
+       - 'zenith,azimuth': (theta, phi) angles commonly used
+       in physics or the (colatitude, longitude) angles used
+       in the celestial and geographical coordinate systems
+       - 'azimuth,zenith': (longitude, colatitude) convention
+       - 'elevation,azimuth: (latitude, longitude) convention
+       - 'azimuth,elevation': (longitude, latitude) convention
+
+    """
+    def __init__(self, convention, degrees=False, **keywords):
+        """
+        convention : string
+            One of the following spherical coordinate conventions:
+            'zenith,azimuth', 'azimuth,zenith', 'elevation,azimuth' and
+            'azimuth,elevation'.
+        degrees : boolean, optional
+            If true, the spherical coordinates are returned in degrees.
+
+        """
+        if degrees:
+            self.__class__ = CompositionOperator
+            self.__init__(
+                [DegreesOperator(),
+                 Cartesian2SphericalOperator(convention, **keywords)])
+            return
+        self.degrees = False
+
+        _CartesianSpherical.__init__(
+            self, convention,
+            reshapein=self._reshapecartesian,
+            reshapeout=self._reshapespherical,
+            validatein=self._validatecartesian,
+            validateout=self._validatespherical,
+            **keywords)
+        self.set_rule('I',
+                      lambda s: Spherical2CartesianOperator(s.convention))
+        self.set_rule(('.', Spherical2CartesianOperator), self._rule_identity,
+                      CompositionOperator)
+
+    def direct(self, input, output):
+        if self.convention.startswith('azimuth'):
+            o1, o2 = output[..., 1], output[..., 0]
+        else:
+            o1, o2 = output[..., 0], output[..., 1]
+        np.arccos(input[..., 2], o1)
+        if 'elevation' in self.convention:
+            np.subtract(pi(self.dtype) / 2, o1, o1)
+        np.arctan2(input[..., 1], input[..., 0], o2)
+        if o2.ndim > 0:
+            o2[o2 < 0] += 2 * pi(self.dtype)
+        elif o2 < 0:
+            o2 += 2 * pi(self.dtype)
+
+
+class Spherical2CartesianOperator(_CartesianSpherical):
+    """
+    Convert spherical coordinates in radians or degrees into unit cartesian
+    vectors.
+
+    The spherical coordinate system is defined by:
+       - the zenith direction of coordinate (0, 0, 1)
+       - the azimuthal reference of coordinate (1, 0, 0)
+       - the azimuth signedness: it is counted positively from the X axis
+    to the Y axis.
+
+    The last dimension of the operator's input is 2 and it encodes
+    the two spherical angles. Four conventions define what these angles are:
+       - 'zenith,azimuth': (theta, phi) angles commonly used
+       in physics or the (colatitude, longitude) angles used
+       in the celestial and geographical coordinate systems
+       - 'azimuth,zenith': (longitude, colatitude) convention
+       - 'elevation,azimuth: (latitude, longitude) convention
+       - 'azimuth,elevation': (longitude, latitude) convention
+
+    """
+    def __init__(self, convention, degrees=False, **keywords):
+        """
+        convention : string
+            One of the following spherical coordinate conventions:
+            'zenith,azimuth', 'azimuth,zenith', 'elevation,azimuth' and
+            'azimuth,elevation'.
+        degrees : boolean, optional
+            If true, the input spherical coordinates are assumed to be in
+            degrees.
+
+        """
+        if degrees:
+            self.__class__ = CompositionOperator
+            self.__init__(
+                [Spherical2CartesianOperator(convention, **keywords),
+                 RadiansOperator()])
+            return
+        self.degrees = False
+
+        _CartesianSpherical.__init__(
+            self, convention,
+            reshapein=self._reshapespherical,
+            reshapeout=self._reshapecartesian,
+            validatein=self._validatespherical,
+            validateout=self._validatecartesian,
+            **keywords)
+        self.set_rule('I',
+                      lambda s: Cartesian2SphericalOperator(s.convention))
+        self.set_rule(('.', Cartesian2SphericalOperator), self._rule_identity,
+                      CompositionOperator)
+
+    def direct(self, input, output):
+        if self.convention.startswith('azimuth'):
+            theta, phi = input[..., 1], input[..., 0]
+        else:
+            theta, phi = input[..., 0], input[..., 1]
+        if 'elevation' in self.convention:
+            theta = 0.5 * pi(self.dtype) - theta
+        sintheta = np.sin(theta)
+        np.multiply(sintheta, np.cos(phi), output[..., 0])
+        np.multiply(sintheta, np.sin(phi), output[..., 1])
+        np.cos(theta, output[..., 2])
+
 
 @square
 @inplace
@@ -44,8 +224,8 @@ class ClipOperator(Operator):
 
     Exemples
     --------
-    >>>  C = ClipOperator(0, 1)
-    >>> x = linspace(-2, 2, 5)
+    >>> C = ClipOperator(0, 1)
+    >>> x = np.linspace(-2, 2, 5)
     >>> x
     array([-2., -1.,  0.,  1.,  2.])
     >>> C(x)
@@ -57,7 +237,7 @@ class ClipOperator(Operator):
 
     """
     def __init__(self, vmin, vmax, **keywords):
-        Operator.__init__(self, lambda i,o: np.clip(i, vmin, vmax, out=o),
+        Operator.__init__(self, lambda i, o: np.clip(i, vmin, vmax, out=o),
                           **keywords)
 
 
@@ -115,7 +295,7 @@ class MaxOperator(ReductionOperator):
 
     """
     def __init__(self, axis=None, dtype=None, skipna=False, **keywords):
-        if np.__version__ < '1.8':
+        if np.__version__ < '2':
             func = np.nanmax if skipna else np.max
         else:
             func = np.max
@@ -148,7 +328,7 @@ class MinOperator(ReductionOperator):
 
     """
     def __init__(self, axis=None, dtype=None, skipna=False, **keywords):
-        if np.__version__ < '1.8':
+        if np.__version__ < '2':
             func = np.nanmin if skipna else np.min
         else:
             func = np.min
@@ -209,7 +389,7 @@ class MaximumOperator(Operator):
     Exemple
     -------
     >>> M = MaximumOperator(1)
-    >>> x =  linspace(-2, 2, 5)
+    >>> x = np.linspace(-2, 2, 5)
     >>> x
     array([-2., -1.,  0.,  1.,  2.])
     >>> M(x)
@@ -221,10 +401,11 @@ class MaximumOperator(Operator):
 
     """
     def __init__(self, value, **keywords):
-        Operator.__init__(self, lambda i,o: np.maximum(i, value, o), **keywords)
+        Operator.__init__(self, lambda i, o: np.maximum(i, value, o),
+                          **keywords)
 
 
-@square 
+@square
 @inplace
 @separable
 class MinimumOperator(Operator):
@@ -239,7 +420,7 @@ class MinimumOperator(Operator):
     Exemple
     -------
     >>> M = MinimumOperator(1)
-    >>> x =  linspace(-2, 2, 5)
+    >>> x = np.linspace(-2, 2, 5)
     >>> x
     array([-2., -1.,  0.,  1.,  2.])
     >>> M(x)
@@ -251,7 +432,28 @@ class MinimumOperator(Operator):
 
     """
     def __init__(self, value, **keywords):
-        Operator.__init__(self, lambda i,o: np.minimum(i, value, o), **keywords)
+        Operator.__init__(self, lambda i, o: np.minimum(i, value, o),
+                          **keywords)
+
+
+@square
+@inplace
+class NormalizeOperator(Operator):
+    """
+    Normalize a cartesian vector.
+
+    Example
+    -------
+    >>> n = NormalizeOperator()
+    >>> n([1, 1])
+    array([ 0.70710678,  0.70710678])
+
+    """
+    def __init__(self, dtype=float, **keywords):
+        Operator.__init__(self, dtype=dtype, **keywords)
+
+    def direct(self, input, output):
+        np.divide(input, np.sqrt(np.sum(input**2, axis=-1))[..., None], output)
 
 
 @square
@@ -259,15 +461,15 @@ class MinimumOperator(Operator):
 class NumexprOperator(Operator):
     """
     Return an operator evaluating an expression using numexpr.
-    
+
     Parameters
     ----------
     expr : string
         The numexp expression to be evaluated. It must contain the 'input'
         variable name.
     global_dict : dict
-        A dictionary of global variables that are passed to numexpr's 'evaluate'
-        method.
+        A dictionary of global variables that are passed to numexpr's
+        'evaluate' method.
 
     Example
     -------
@@ -280,9 +482,8 @@ class NumexprOperator(Operator):
     def __init__(self, expr, global_dict=None, dtype=float, **keywords):
         self.expr = expr
         self.global_dict = global_dict
-        if numexpr.__version__ < '2.0.2':
-            keywords['flags'] = self.validate_flags(keywords.get('flags', {}),
-                                                    inplace_reduction=False)
+        if numexpr.__version__ < '2.1':
+            self._disable_inplace_reduction = True
         Operator.__init__(self, dtype=dtype, **keywords)
 
     def direct(self, input, output, operation=operation_assignment):
@@ -299,8 +500,9 @@ class NumexprOperator(Operator):
 @inplace
 @separable
 class RoundOperator(Operator):
-    """Rounding operator.
-    
+    """
+    Rounding operator.
+
     The rounding method may be one of the following:
         - rtz : round towards zero (truncation)
         - rti : round towards infinity (Not implemented)
@@ -310,27 +512,28 @@ class RoundOperator(Operator):
         - rhti : round half towards infinity (Fortran's nint)
         - rhtmi : round half towards minus infinity
         - rhtpi : round half towards positive infinity
-        - rhte : round half to even (numpy's round), 
+        - rhte : round half to even (numpy's round),
         - rhto : round half to odd
         - rhs : round half stochastically (Not implemented)
 
     """
     def __init__(self, method='rhte', **keywords):
         method = method.lower()
-        table = {'rtz'   : np.trunc,
+        table = {'rtz': np.trunc,
                  #'rti'
-                 'rtmi'  : np.floor,
-                 'rtpi'  : np.ceil,
+                 'rtmi': np.floor,
+                 'rtpi': np.ceil,
                  #'rhtz'
                  #'rhti'
-                 'rhtmi' : self._direct_rhtmi,
-                 'rhtpi' : self._direct_rhtpi,
-                 'rhte'  : lambda i,o: np.round(i,0,o),
+                 'rhtmi': self._direct_rhtmi,
+                 'rhtpi': self._direct_rhtpi,
+                 'rhte': lambda i, o: np.round(i, 0, o),
                  #'rhs'
                  }
         if method not in table:
-            raise ValueError('Invalid rounding method. Expected values are {0}.'
-                             .format(strenum(table.keys())))
+            raise ValueError(
+                'Invalid rounding method. Expected values are {0}.'.format(
+                strenum(table.keys())))
         Operator.__init__(self, table[method], **keywords)
         self.method = method
 
@@ -349,7 +552,6 @@ class RoundOperator(Operator):
         np.add(output, 1, output)
 
 
-@real
 @square
 @idempotent
 @inplace
@@ -381,7 +583,7 @@ class HardThresholdingOperator(Operator):
             return
         Operator.__init__(self, **keywords)
         self.a = a
-        self.set_rule('.{HardThresholdingOperator}', lambda s, o:
+        self.set_rule(('.', HardThresholdingOperator), lambda s, o:
                       HardThresholdingOperator(np.maximum(s.a, o.a)),
                       CompositionOperator)
 
@@ -389,7 +591,6 @@ class HardThresholdingOperator(Operator):
         hard_thresholding(input, self.a, output)
 
 
-@real
 @square
 @inplace
 @separable
@@ -424,7 +625,6 @@ class SoftThresholdingOperator(Operator):
         soft_thresholding(input, self.a, output)
 
 
-@real
 @separable
 class _1dNdOperator(Operator):
     """ Base class for 1d-Nd coordinate mappings. """
@@ -494,8 +694,7 @@ class To1dOperator(_1dNdOperator):
         if 'validatein' not in keywords:
             keywords['validatein'] = self._validate_to1d
         _1dNdOperator.__init__(self, shape_, order=order, **keywords)
-        self.set_rule('.I', lambda s: ReverseOperatorFactory(
-            ToNdOperator, s, s.shape_, order=s.order, **keywords))
+        self.set_rule('I', lambda s: ToNdOperator(s.shape_, order=s.order))
 
     def direct(self, input, output):
         np.dot(input, self.coefs, out=output)
@@ -536,8 +735,8 @@ class ToNdOperator(_1dNdOperator):
         if 'validateout' not in keywords:
             keywords['validateout'] = self._validate_to1d
         _1dNdOperator.__init__(self, shape_, order=order, **keywords)
-        self.set_rule('.I', lambda s: ReverseOperatorFactory(
-            To1dOperator, s, s.shape_, order=s.order, **keywords))
+        self.set_rule('I', lambda s: To1dOperator(
+            s.shape_, order=s.order))
 
     def direct(self, input, output):
         np.floor_divide(input[..., None], self.coefs, out=output)
